import threading
import logging
import queue
import concurrent.futures

import pebble

import paho.mqtt.client as mqtt

from citadel.mqtt import Configuration

from .action import Action
from .watch import Watch

class Scene(threading.Thread, mqtt.Client):
    """ A scene run as a thread, has his own mqtt session

    Can be activated by topic:

        cmnd/scene/{mqtt_name}/activate ON|OFF|null

    And return status on topic:

        stat/scene/{mqtt_name}/state  ON|OFF

    Special class attributes:

    mqtt_name: define topic name, default class name in lower case

    """
    TOPIC_CMND = "cmnd/scene/{mqtt_name}/activate"
    TOPIC_STAT = "stat/scene/{mqtt_name}/activate"

    STATE_DEACTIVATE = 0
    STATE_ACTIVATE = 1
    STATE_ACTIVATING = 2

    def __init__(self, config: Configuration, logger: logging.Logger=None):
        threading.Thread.__init__(self)
        mqtt.Client.__init__(self)

        self.configuration = config

        if not hasattr(self, 'mqtt_name'):
            self.mqtt_name = self.__class__.__name__.lower()

        if not logger:
            logger = logging.getLogger('scene.%s' % self.mqtt_name)
        self.logger = logger

        # we need a queue for publishing message from different thread
        self.__pub_queue = queue.Queue()
        self.__state = Scene.STATE_DEACTIVATE

        self.__executor = None
        self.__activating_r_futures = []

    @property
    def scene_state(self):
        return self.__state

    def run(self):
        self.username_pw_set(self.configuration.user,
                self.configuration.password)
        self.connect(self.configuration.host,
                self.configuration.port)

        self.subscribe(self.TOPIC_CMND.format(mqtt_name=self.mqtt_name), 2)

        for watch in self.WATCH:
            self.subscribe(watch.topic, 1)

        run = True
        while run:
            #Check publish queue
            try:
                (args, kwargs) = self.__pub_queue.get(block=False)
                self.publish(*args, **kwargs)
            except queue.Empty:
                pass

            # Check for futures if activating
            if self.scene_state == Scene.STATE_ACTIVATING:
                for (action, futures) in self.__activating_r_futures:
                    try:
                        r = futures.result(timeout=0)
                        if r != Action.SUCCESS:
                            self.logger.error("Action: '%s' error", action.name)
                            self.logger.info('Activating error')
                            self.__executor.join()
                            self.__executor = None
                            self.__activating_r_futures = []
                            self.__state = Scene.STATE_DEACTIVATE
                            self.__send_state()
                            break
                    except concurrent.futures.TimeoutError:
                        break
                else:
                    self.logger.info('Activating terminate')
                    self.__executor.join()
                    self.__executor = None
                    self.__activating_r_futures = []
                    self.__state = Scene.STATE_ACTIVATE
                    self.__send_state()

            #Mqtt client loop
            self.loop()


    def threaded_publish(self, *args, **kwargs):
        self.__pub_queue.put((args, kwargs))

    def on_message(self, mqttc, obj, msg):
        payload = msg.payload.decode()
        topic = msg.topic

        # on CMND
        if topic == self.TOPIC_CMND.format(mqtt_name=self.mqtt_name):
            if payload.lower() == 'on':
                self.logger.info('CMND activate ON')
                self.__activate()
            elif payload.lower() == 'off':
                self.logger.info('CMND activate OFF')
                self.__deactivate()
            elif payload == '':
                self.logger.info('CMND activate state')
                self.__send_state()
            else:
                self.logger.warn('CMND activate invalide')

        elif self.scene_state == Scene.STATE_ACTIVATE:
            for watch in self.WATCH:
                if topic == watch.topic and not watch.check(topic, payload):
                    self.logger.info('Watched topic %s with invalid payload %s deactivating', topic, payload)
                    self.__deactivate()
                    break

    def __activate(self):
        if self.scene_state:
            self.logger.info('Already activate/activating ignore cmnd')
            return
        if self.__executor != None:
            self.logger.error('Trying to activate but __executor not None')
            return

        self.__state = Scene.STATE_ACTIVATING
        self.__executor = pebble.ThreadPool()

        for a in self.ON:
            self.__activating_r_futures.append(
                    (a, self.__executor.schedule(a.setup(self).run))
            )

        self.__executor.close()
        self.__send_state()

    def __deactivate(self):
        self.__state = Scene.STATE_DEACTIVATE
        self.__send_state()

    def __send_state(self):
        self.publish(self.TOPIC_STAT.format(mqtt_name=self.mqtt_name),
                payload='ON' if self.scene_state else 'OFF',
                qos=2,
                retain=False)
